/*
 * Java 4K 2012
 * Copyright (C) 2012 meatfighter.com
 *
 * This file is part of Java 4K 2012.
 *
 * Java 4K 2012 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java 4K 2012 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

import java.applet.*;
import java.awt.*;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.image.*;
import javax.imageio.*;
import java.util.*;

public class Main extends Applet implements Runnable {
  
  public static final int APPLET_WIDTH = 640;
  public static final int APPLET_HEIGHT = 420;
  
  public static final float PLAYER_SPEED = 0.75f;
  public static final float BULLET_SPEED = 3f;
  public static final float BULLET_DIAGONAL_SPEED 
      = (float)(BULLET_SPEED / Math.sqrt(2));
  
  private enum PlayerState {
    WALKING,
    RESTING,
    SHOOTING,
    HIT,
  }
  
  private class Player {
    public int x;
    public int y;
    public int minX;
    public int minY;
    public int maxX;
    public int maxY;
    public BufferedImage sprite;
    public PlayerState state;
    public int countDown;
    public float fx;
    public float fy;
    public float vx;
    public float vy;
    public float bulletX;
    public float bulletY;
    public float bulletVx;
    public float bulletVy;
    public boolean shot;
    public Color bulletColor;
  }
  
  private BufferedImage man0walk0;
  private BufferedImage man0walk1;
  private BufferedImage man0shootUp;
  private BufferedImage man0shootHorizontal;
  private BufferedImage man0shootDown;
  private BufferedImage man0dead;
  private BufferedImage man1walk0;
  private BufferedImage man1walk1;
  private BufferedImage man1shootUp;
  private BufferedImage man1shootHorizontal;
  private BufferedImage man1shootDown;
  private BufferedImage man1dead;
  private BufferedImage[] digits = new BufferedImage[10];
  
  private boolean[][] title = new boolean[19][23];
  private boolean[][] titleCopy = new boolean[19][23];
  
  private Color BACKGROUND_COLOR = new Color(0x2D32B8);
  private Color TITLE_COLOR = new Color(0xA459D0);
  private Color PLAYER0_COLOR = new Color(0xB47A30);
  private Color PLAYER1_COLOR = new Color(0xD65C5C);
  
  private int man0Score = 0;
  private int man1Score = 0;
  private Player[] players = new Player[2];
  
  private Random random = new Random();
  
  private int appletWidth = 640;
  private int appletHeight = 420;
  
  private void loadSprites() {
    try {
      man0walk0 = ImageIO.read(Main.class.getResource("/images/0.png"));
      man0walk1 = ImageIO.read(Main.class.getResource("/images/1.png"));
      man0shootUp = ImageIO.read(Main.class.getResource("/images/2.png"));
      man0shootHorizontal 
          = ImageIO.read(Main.class.getResource("/images/3.png"));
      man0shootDown = ImageIO.read(Main.class.getResource("/images/4.png"));
      man0dead = ImageIO.read(Main.class.getResource("/images/5.png"));
      man1walk0 = ImageIO.read(Main.class.getResource("/images/6.png"));
      man1walk1 = ImageIO.read(Main.class.getResource("/images/7.png"));
      man1shootUp = ImageIO.read(Main.class.getResource("/images/8.png"));
      man1shootHorizontal 
          = ImageIO.read(Main.class.getResource("/images/9.png"));
      man1shootDown = ImageIO.read(Main.class.getResource("/images/10.png"));
      man1dead = ImageIO.read(Main.class.getResource("/images/11.png"));
      
      BufferedImage titleImage = ImageIO.read(
          Main.class.getResource("/images/12.png"));
      for(int y = 0; y < 17; y++) {
        for(int x = 0; x < 21; x++) {
          titleCopy[y + 1][x + 1] = title[y + 1][x + 1] 
              = (titleImage.getRGB(x, y) & 1) == 1;
        }
      }
      
      for(int i = 0; i < 10; i++) {
        digits[i] = ImageIO.read(Main.class.getResource("/images/" 
            + (13 + i) + ".png"));
      }

    } catch(Throwable t) {
      t.printStackTrace();
    }
  }
  
  private void initPlayers() {
    players[0] = new Player();
    players[1] = new Player();
    
    players[0].bulletColor = PLAYER0_COLOR;
    players[1].bulletColor = PLAYER1_COLOR;
    
    players[0].fy = players[1].fy = 136;
    players[0].fx = 28;
    players[1].fx = 544;
    players[0].sprite = man0walk0;
    players[1].sprite = man1walk0;
    
    players[0].minX = 16;
    players[0].minY = 80;
    players[0].maxX = 128;
    players[0].maxY = 274;
    
    players[1].minX = 448;
    players[1].minY = 80;
    players[1].maxX = 560;
    players[1].maxY = 274;
    
    players[0].state = players[1].state = PlayerState.RESTING;
  }

  @Override
  public void start() {
    new Thread(this).start();
    
    addComponentListener(new ComponentListener() {
      @Override
      public void componentResized(ComponentEvent e) {
        appletResized();
      }
      @Override
      public void componentMoved(ComponentEvent e) {
      }
      @Override
      public void componentShown(ComponentEvent e) {
      }
      @Override
      public void componentHidden(ComponentEvent e) {
      }
    });
  }
  
  private void appletResized() {
    appletWidth = getWidth();
    appletHeight = getHeight();
  }
  
  public void run() {

    loadSprites();
    initPlayers();

    BufferedImage image = new BufferedImage(APPLET_WIDTH, APPLET_HEIGHT, 
        BufferedImage.TYPE_INT_RGB);
    Graphics2D g = (Graphics2D)image.getGraphics();
    Graphics2D g2 = null;

    long nextFrameStartTime = System.nanoTime();
    while(true) {

      do {
        nextFrameStartTime += 16666667;

        update();

      } while(nextFrameStartTime < System.nanoTime());

      render(g);

      // show the hidden buffer
      if (g2 == null) {
        g2 = (Graphics2D)getGraphics();        
        requestFocus();
        EventQueue.invokeLater(new Runnable() {
          @Override
          public void run() {
            appletResized();
          }
        });
      } else {
        g2.drawImage(image, 0, 0, appletWidth, appletHeight, null);
      }

      // burn off extra cycles
      while(nextFrameStartTime - System.nanoTime() > 0) {
        Thread.yield();
      }
    }
  } 
  
  private void playerRests(Player player) {
    player.state = PlayerState.RESTING;
    player.countDown = random.nextInt(30);
  }
  
  private void playerWalks(Player player, int index) {
    player.state = PlayerState.WALKING;
    float targetX = (player.maxX - player.minX) * random.nextFloat() 
        + player.minX;
    float targetY = (player.maxY - player.minY) * random.nextFloat() 
        + player.minY;
    player.vx = targetX - player.fx;
    player.vy = targetY - player.fy;
    float mag = (float)Math.hypot(player.vx, player.vy);
    player.countDown = (int)(mag / PLAYER_SPEED);    
    float k = PLAYER_SPEED / mag;
    player.vx *= k;
    player.vy *= k;
    if (index == 0) {
      if (player.sprite != man0walk0 && player.sprite != man0walk1) {
        player.sprite = man0walk0;
      }
    } else {
      if (player.sprite != man1walk0 && player.sprite != man1walk1) {
        player.sprite = man1walk0;
      }
    }
  }
  
  private void playerShoots(Player player, int index) {
    player.state = PlayerState.SHOOTING;
    player.shot = true;
    int direction = random.nextInt(3);
    switch(direction) {
      case 0:
        if (index == 0) {
          player.sprite = man0shootUp;
          player.bulletX = player.fx + 48;
          player.bulletY = player.fy + 32;
          player.bulletVx = BULLET_DIAGONAL_SPEED;
          player.bulletVy = -BULLET_DIAGONAL_SPEED;
        } else {
          player.sprite = man1shootUp;
          player.bulletX = player.fx + 8;
          player.bulletY = player.fy + 32;
          player.bulletVx = -BULLET_DIAGONAL_SPEED;
          player.bulletVy = -BULLET_DIAGONAL_SPEED;
        }
        break;
      case 1:
        if (index == 0) {
          player.sprite = man0shootHorizontal;
          player.bulletX = player.fx + 56;
          player.bulletY = player.fy + 40;
          player.bulletVx = BULLET_SPEED;
          player.bulletVy = 0;
        } else {
          player.sprite = man1shootHorizontal;
          player.bulletX = player.fx + 0;
          player.bulletY = player.fy + 40;
          player.bulletVx = -BULLET_SPEED;
          player.bulletVy = 0;
        }
        break;
      case 2:
        if (index == 0) {
          player.sprite = man0shootDown;
          player.bulletX = player.fx + 48;
          player.bulletY = player.fy + 64;
          player.bulletVx = BULLET_DIAGONAL_SPEED;
          player.bulletVy = BULLET_DIAGONAL_SPEED;
        } else {
          player.sprite = man1shootDown;
          player.bulletX = player.fx + 8;
          player.bulletY = player.fy + 64;
          player.bulletVx = -BULLET_DIAGONAL_SPEED;
          player.bulletVy = BULLET_DIAGONAL_SPEED;
        }
        break;
    }
  }
  
  private void update() {
    for(int i = 0; i < 2; i++) {
      Player player = players[i];  
      if (player.shot) {
        boolean done = false;
        
        for(int j = 0; j < 2 && !done; j++) {
          player.bulletX += player.bulletVx;
          player.bulletY += player.bulletVy;

          if (player.bulletY <= 80) {
            player.bulletVy = Math.abs(player.bulletVy);
          } else if (player.bulletY >= 370) {
            player.bulletVy = -Math.abs(player.bulletVy);
          }
          
          boolean resetTitle = false;
          try {
            if (i == 0) {
              if (players[1].state != PlayerState.HIT
                  && player.bulletX >= players[1].fx - 4
                  && player.bulletY >= players[1].fy - 4
                  && player.bulletX < players[1].fx + 60
                  && player.bulletY < players[1].fy + 92
                  && players[1].sprite.getRGB(
                    ((int)(player.bulletX + 4 - players[1].fx)) >> 3,
                    ((int)(player.bulletY + 4 - players[1].fy)) >> 3) < 0) { 
                done = true;
                players[1].state = PlayerState.HIT;
                players[1].sprite = man1dead;
                players[1].countDown = 120;
                man0Score++;
                resetTitle = true;
              }
            } else {
              if (players[0].state != PlayerState.HIT
                  && player.bulletX >= players[0].fx - 4
                  && player.bulletY >= players[0].fy - 4
                  && player.bulletX < players[0].fx + 60
                  && player.bulletY < players[0].fy + 92
                  && players[0].sprite.getRGB(
                    ((int)(player.bulletX + 4 - players[0].fx)) >> 3,
                    ((int)(player.bulletY + 4 - players[0].fy)) >> 3) < 0) { 
                done = true;
                players[0].state = PlayerState.HIT;
                players[0].sprite = man0dead;
                players[0].countDown = 120;
                man1Score++;
                resetTitle = true;
              }
            }
          } catch(Throwable t) {            
          }
          if (resetTitle) {
            for(int y = 0; y < 19; y++) {
              for(int x = 0; x < 23; x++) {
                title[y][x] = titleCopy[y][x];
              }
            }
          }

          if (player.bulletX <= 16 || player.bulletX >= 624) {
            done = true;
          } else if (player.bulletX >= 228 && player.bulletY >= 144
              && player.bulletX <= 404 && player.bulletY <= 288) {
            int x = ((int)(player.bulletX - 224)) >> 3;
            int y = ((int)(player.bulletY - 140)) >> 3;
            if (!title[y][x]) {
              title[y][x] = true;
              done = true;
            }
          }

          if (done) {
            player.shot = false;
            if (player.state != PlayerState.HIT) {
              player.state = PlayerState.RESTING;
              player.countDown = 5 + random.nextInt(10);
              player.sprite = i == 0 ? man0walk0 : man1walk0;
            }
          }
        }        
      } else if (--player.countDown <= 0) {
        switch(random.nextInt(3)) {
          case 0:
            playerWalks(player, i);
            break;
          case 1:
            playerRests(player);
            break;
          case 2:
            playerShoots(player, i);
            break;
        }        
      } else {
        switch(player.state) {
          case RESTING:
            break;
          case WALKING:
            player.fx += player.vx;
            player.fy += player.vy;
            if ((player.countDown & 15) == 0) {
              if (i == 0) {
                player.sprite = (player.sprite == man0walk0) 
                    ? man0walk1 : man0walk0;
              } else {
                player.sprite = (player.sprite == man1walk0) 
                    ? man1walk1 : man1walk0;
              }
            }
            break;
        }
      }
      player.x = (int)player.fx;
      player.y = (int)player.fy;
    }
  }
  
  private void drawScore(Graphics2D g, int score, int x, int y) {
    do {
      g.drawImage(digits[score % 10], x, y, 48, 20, null);
      x -= 48;
      score /= 10;
    } while(score > 0);
  }
  
  private void render(Graphics2D g) {
    
    // draw solid background
    g.setColor(BACKGROUND_COLOR);
    g.fillRect(0, 0, APPLET_WIDTH, APPLET_HEIGHT);
    
    // draw borders
    g.setColor(TITLE_COLOR);
    g.fillRect(0, 64, APPLET_WIDTH, 16);
    g.fillRect(0, 370, APPLET_WIDTH, 16);
    g.fillRect(0, 80, 16, 290);
    g.fillRect(624, 80, 16, 290);
    
    // draw title    
    for(int y = 0; y < 19; y++) {
      for(int x = 0; x < 23; x++) {
        if (!title[y][x]) {
          g.fillRect(228 + (x << 3), 144 + (y << 3), 8, 8);
        }
      }
    }
    
    // draw players
    for(int i = 0; i < 2; i++) {
      Player player = players[i];
      g.drawImage(player.sprite, player.x, player.y, 64, 96, null);
      if (player.shot) {
        g.setColor(player.bulletColor);
        g.fillRect((int)player.bulletX, (int)player.bulletY, 8, 8);
      }
    }
    
    // draw scores
    drawScore(g, man0Score, 128, 16);
    drawScore(g, man1Score, 448, 16);
  }
  
  public static void main(String[] args) throws Throwable {
    javax.swing.JFrame frame = new javax.swing.JFrame("Java 4K 2012");
    frame.setDefaultCloseOperation(javax.swing.JFrame.EXIT_ON_CLOSE);
    Main applet = new Main();
    applet.setPreferredSize(
        new java.awt.Dimension(APPLET_WIDTH, APPLET_HEIGHT));
    frame.add(applet, java.awt.BorderLayout.CENTER);
    //frame.setResizable(false);
    frame.pack();
    frame.setLocationRelativeTo(null);
    frame.setVisible(true);
    Thread.sleep(250);
    applet.start();
  }  
}
